# frozen_string_literal: true

require_relative "canistor/version"
require_relative "canistor/storage"
require_relative "canistor/authorization"
require_relative "canistor/subject"
require_relative "canistor/error_handler"
require_relative "canistor/handler"
require_relative "canistor/plugin"

require "thread"

# Replacement for the HTTP handler in the AWS SDK that mocks all interaction
# with S3 just above the HTTP level.
#
# The mock implementation is turned on by removing the NetHttp handlers that
# comes with the library by the Canistor handler.
#
#   Aws::S3::Client.remove_plugin(Seahorse::Client::Plugins::NetHttp)
#   Aws::S3::Client.add_plugin(Canistor::Plugin)
#
# The Canistor instance then needs to be configured with buckets and
# credentials to be useful. It can be configured using either the
# config method on the instance or by specifying the buckets one by one.
#
# In the example below Canistor will have two accounts and four buckets. It
# also specifies which accounts can access the buckets.
#
#   Canistor.config(
#     logger: Rails.logger,
#     credentials: {
#       {
#         access_key_id: 'AKIAIXXXXXXXXXXXXXX1',
#         secret_access_key: 'phRL+xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx1'
#       },
#       {
#         access_key_id: 'AKIAIXXXXXXXXXXXXXX2',
#         secret_access_key: 'phRL+xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx2'
#       }
#     },
#     buckets: {
#       'us-east-1' => {
#         'io-canistor-production-images' => {
#           allow_access_keys: ['AKIAIXXXXXXXXXXXXXX1'],
#           replicate_to: [
#             'eu-central-1:io-canistor-production-images-replica'
#           ],
#           versioned: true
#         },
#         'io-canistor-production-books' => {
#           allow_access_keys: ['AKIAIXXXXXXXXXXXXXX1', 'AKIAIXXXXXXXXXXXXXX2']
#         }
#       },
#       'eu-central-1' => {
#         'io-canistor-production-sales' => {
#           allow_access_keys: ['AKIAIXXXXXXXXXXXXXX1']
#         },
#         'io-canistor-production-images-replica' => {
#           allow_access_keys: ['AKIAIXXXXXXXXXXXXXX1'],
#           versioned: true
#         }
#       }
#     }
#   )
#
# Canistor implements basic interaction with buckets and objects. It also
# verifies authentication information. It does not implement control lists so
# all accounts have full access to the buckets and objects.
#
# It's possible to turn on replication and versioning. Note that replication is
# instant in the mock. On actual S3 it takes a while for objects to replicate.
# Not the entire replication and versioning API is implemented.
# 
# The mock can simulate a number of failures. These are triggered by setting
# the operation which needs to fail on the mock. For more information see
# [Canistor.fail].
#
# In most cases you should configure the suite to clear the mock before running 
# each example with [Canistor.clear].
module Canistor
  # Raised when Canistor's own configuration is somehow unusable.
  class ConfigurationError < StandardError; end

  class << self
    attr_accessor :logger
    attr_accessor :store
    attr_reader :credentials
    attr_reader :fail
    attr_reader :fail_mutex
  end

  @store = {}
  @buckets = {}
  @credentials = Set.new
  @fail_mutex = Mutex.new

  def self.find_credentials(authorization)
    if authorization.access_key_id
      credentials.each do |attributes|
        if authorization.access_key_id == attributes[:access_key_id]
          return Aws::Credentials.new(
            attributes[:access_key_id],
            attributes[:secret_access_key]
          )
        end
      end
    end
    nil
  end

  def self.find_bucket(region, bucket)
    store.dig(region, bucket) || find_bucket_by_name_and_warn(region, bucket)
  end

  def self.find_bucket_by_name_and_warn(region, bucket)
    found = find_bucket_by_name(bucket)
    return if found.nil?

    logger.info(
      "S3 client configured for \"#{region}\" but the bucket \"#{bucket}\" " \
      "is in \"#{found.region}\"; Please configure the proper region to " \
      "avoid multiple unnecessary redirects and signing attempts"
    ) if logger
    found
  end

  def self.find_bucket_by_name(bucket)
    store.each do |_, buckets|
      if found = buckets[bucket]
        return found
      end
    end
    nil
  end

  def self.credentials=(accounts)
    accounts.each do |attributes|
      unless attributes.keys.map(&:to_s) == %w(access_key_id secret_access_key)
        raise(
          ArgumentError,
          "Credentials need to specify access_key_id and secret_access_key, " \
          "got: `#{attributes.keys.inspect}'"
        )
      end
    end
    credentials.merge(accounts)
  end

  def self.buckets=(buckets)
    buckets.each do |region, attributes|
      attributes.each do |bucket, options|
        bucket = create_bucket(region, bucket)
        bucket.update_settings(options)
        bucket
      end
    end
  end

  # Configures a bucket in the mock implementation. Use #update_settings on
  # the Container object returned by this method to configure who may access
  # the bucket.
  def self.create_bucket(region, bucket_name)
    store[region] ||= {}
    store[region][bucket_name] = Canistor::Storage::Bucket.new(
      region: region,
      name: bucket_name
    )
  end

  def self.config(config)
    config.each do |section, attributes|
      public_send("#{section}=", attributes)
    end
  end

  SUPPORTED_FAILURES = [
    :internal_server_error,
    :reset_connection,
    :bad_request,
    :fatal,
    :store
  ]

  # The mock can simulate a number of failures. These are triggered by setting
  # the way we expect it to fail. Note that the AWS-SDK already helps you
  # to recover from certain errors like :reset_connection. If you want these
  # kinds of error to trigger a failure you have to call #fail more than then
  # configured retry count.
  #
  #   Canistor.fail(:reset_connection)
  #   Canistor.fail(
  #     :reset_connection,
  #     :reset_connection,
  #     :reset_connection,
  #     :reset_connection
  #   )
  #
  #  * reset_connection: Signals the library to handle a connection error
  #    (retryable)
  #  * internal_server_error: Returns a 500 internal server error (retryable)
  #  * bad_request: Returns a 400 bad request (not retryable)
  #  * fatal: Signals the library to handle a fatal error (fatal)
  #
  # A less common problem is when S3 reports a successful write but fails to
  # store the file. This means the PUT to the bucket will be successful, but
  # GET and HEAD on the object fail, because it's not there.
  #
  #   Canistor.fail(:store)
  def self.fail(*operations)
    unsupported = operations - SUPPORTED_FAILURES
    unless unsupported.empty?
      raise(
        ArgumentError,
        "Requested unsupported failure: `#{unsupported.inspect}', supported: " \
        "#{SUPPORTED_FAILURES.inspect}."
      )
    end
    fail_mutex.synchronize do
      @fail.concat(operations)
    end
  end

  # Executes the block when the operation is in the failure queue and removes
  # one instance of the operation.
  def self.take_fail(operation, &block)
    fail_mutex.synchronize do
      if index = @fail.index(operation)
        begin
          block.call
        ensure
          @fail.delete_at(index)
        end
      end
    end
  end

  # Clears the state of the mock. Leaves all the credentials and buckets but
  # removes all objects and mocked responses.
  def self.clear
    @fail = []
    @store.each do |region, buckets|
      buckets.each do |bucket_name, bucket|
        bucket.clear
      end
    end
  end

  clear
end
