# frozen_string_literal: true

module Canistor
  class Authorization
    attr_reader :protocol
    attr_reader :region
    attr_reader :access_key_id
    attr_reader :signature
    attr_reader :date

    def initialize(authorization)
      @protocol, params = authorization.split(' ', 2)
      params.split(', ').inject({}) do |unpacked, part|
        name, values = part.split('=')
        case name
        when 'Credential'
          self.credential = values.split('/')
        when 'Signature'
          self.signature = values
        end
        unpacked
      end
    end

    def valid_signature?(request, credentials)
      return false if signature.to_s.strip == ''
      signer = Aws::Sigv4::Signer.new(
        service: 's3',
        region: region,
        credentials_provider: credentials,
        uri_escape_path: false,
        unsigned_headers: ['content-length', 'x-amzn-trace-id']
      )
      signed_request = signer.sign_request(
        http_method: request.http_method,
        url: request.endpoint.to_s,
        headers: request.headers.to_hash,
        body: request.body
      )
      signature == signer.send(
        :signature,
        credentials.secret_access_key,
        date,
        signed_request.string_to_sign
      )
    end

    private

    def signature=(signature)
      @signature = signature
    end

    def credential=(credential)
      @access_key_id, @date, @region = credential[0, 3]
    end
  end
end
