# frozen_string_literal: true

module Canistor
  class ErrorHandler
    attr_reader :context
    attr_reader :request_id
    attr_reader :host_id

    def initialize(context)
      @context = context
      @request_id = SecureRandom.hex(8).upcase
      @host_id = Base64.strict_encode64(SecureRandom.hex(16))
    end

    def request
      context.http_request
    end

    def response
      context.http_response
    end

    def serve_continue
      response.signal_headers(
        100,
        'date' => Time.now.httpdate,
        'x-amz-request-id' => request_id
      )
    end

    def serve_bad_request
      serve_error(400, Nokogiri::XML::Builder.new do |xml|
        xml.Error do
          xml.Code 'InvalidArgument'
          xml.Message 'Invalid Argument.'
          xml.RequestId request_id
          xml.HostId host_id
        end
      end.to_xml)
    end

    def serve_invalid_access_key(authorization)
      serve_error(403, Nokogiri::XML::Builder.new do |xml|
        xml.Error do
          xml.Code 'InvalidAccessKeyId'
          xml.Message 'The AWS Access Key Id you provided does not exist in our records.'
          xml.AWSAccessKeyId authorization.access_key_id
          xml.RequestId request_id
          xml.HostId host_id
        end
      end.to_xml)
    end

    def serve_signature_does_not_match(authorization)
      serve_error(403, Nokogiri::XML::Builder.new do |xml|
        xml.Error do
          xml.Code 'SignatureDoesNotMatch'
          xml.Message 'The request signature we calculated does not match the signature you provided. Check your key and signing method.'
          xml.AWSAccessKeyId authorization.access_key_id
          xml.SignatureProvided authorization.signature
          xml.RequestId request_id
          xml.HostId host_id
        end
      end.to_xml)
    end

    def serve_no_such_bucket(subject)
      serve_error(404, Nokogiri::XML::Builder.new do |xml|
        xml.Error do
          xml.Code 'NoSuchBucket'
          xml.Message 'The specified bucket does not exist'
          xml.BucketName subject.bucket
          xml.RequestId request_id
          xml.HostId host_id
        end
      end.to_xml)
    end

    def serve_no_such_upload(subject)
      serve_error(404, Nokogiri::XML::Builder.new do |xml|
        xml.Error do
          xml.Code 'NoSuchUpload'
          xml.Message 'The specified upload does not exist. The upload ID ' \
                      'may be invalid, or the upload may have been aborted ' \
                      'or completed.'
          xml.Key subject.key
          xml.RequestId request_id
          xml.HostId host_id
        end
      end.to_xml)
    end

    def serve_no_such_key(subject)
      serve_error(404, Nokogiri::XML::Builder.new do |xml|
        xml.Error do
          xml.Code 'NoSuchKey'
          xml.Message 'The specified key does not exist.'
          xml.Key subject.key
          xml.RequestId request_id
          xml.HostId host_id
        end
      end.to_xml)
    end

    def serve_invalid_part(upload_id, part_number, part_etag)
      serve_error(400, Nokogiri::XML::Builder.new do |xml|
        xml.Error do
          xml.Code 'InvalidPart'
          xml.Message 'One or more of the specified parts could not be found. '\
                      'The part may not have been uploaded, or the specified ' \
                      ' entity tag may not match the part\'s entity tag.'
          xml.UploadId upload_id
          xml.PartNumber part_number
          xml.ETag part_etag
          xml.RequestId request_id
          xml.HostId host_id
        end
      end.to_xml)
    end

    def serve_invalid_part_order(upload_id)
      serve_error(400, Nokogiri::XML::Builder.new do |xml|
        xml.Error do
          xml.Code 'InvalidPartOrder'
          xml.Message 'The list of parts was not in ascending order. Parts ' \
                      'must be ordered by part number.'
          xml.UploadId upload_id
          xml.RequestId request_id
          xml.HostId host_id
        end
      end.to_xml)
    end

    def serve_access_denied(subject)
      serve_error(403, Nokogiri::XML::Builder.new do |xml|
        xml.Error do
          xml.Code 'AccessDenied'
          xml.Message 'Access Denied'
          xml.RequestId request_id
          xml.HostId host_id
        end
      end.to_xml)
    end

    def serve_internal_error
      serve_error(500, Nokogiri::XML::Builder.new do |xml|
        xml.Error do
          xml.Code 'InternalError'
          xml.Message 'We encountered an internal error. Please try again.'
          xml.RequestId request_id
          xml.HostId host_id
        end
      end.to_xml)
    end

    def serve_error(status_code, body)
      response.signal_headers(
        status_code,
        'date' => Time.now.httpdate,
        'x-amz-request-id' => request_id
      )
      unless request.http_method == 'HEAD'
        response.signal_data(body)
      end
    end

    def trigger_reset_connection
      response.signal_error(Seahorse::Client::NetworkingError.new(
        Errno::ECONNRESET.new, 'Remote host reset the connection request.'
      ))
    end

    def trigger_fatal_error
      response.signal_error(RuntimeError.new("Fatal error."))
    end

    def self.serve_continue(context)
      new(context).serve_continue
    end

    def self.serve_bad_request(context)
      new(context).serve_bad_request
    end

    def self.serve_invalid_access_key(context, authorization)
      new(context).serve_invalid_access_key(authorization)
    end

    def self.serve_signature_does_not_match(context, authorization)
      new(context).serve_signature_does_not_match(authorization)
    end

    def self.serve_no_such_bucket(context, subject)
      new(context).serve_no_such_bucket(subject)
    end

    def self.serve_no_such_upload(context, subject)
      new(context).serve_no_such_upload(subject)
    end

    def self.serve_no_such_key(context, subject)
      new(context).serve_no_such_key(subject)
    end

    def self.serve_invalid_part(context, upload_id, part_number, part_etag)
      new(context).serve_invalid_part(upload_id, part_number, part_etag)
    end

    def self.serve_invalid_part_order(context, upload_id)
      new(context).serve_invalid_part_order(upload_id)
    end

    def self.access_denied(context, subject)
      new(context).access_denied(subject)
    end

    def self.serve_internal_error(context)
      new(context).serve_internal_error
    end

    def self.trigger_reset_connection(context)
      new(context).trigger_reset_connection
    end

    def self.trigger_fatal_error(context)
      new(context).trigger_fatal_error
    end
  end
end
