# frozen_string_literal: true

module Canistor
  module Storage
    class Bucket
      # Stores settings for the bucket.
      #
      #   +allow_access_keys+: Allow full access to the bucket using these
      #   access keys.
      #   +replicate_to+: Replicate all updates to the bucket to this bucket.
      #   +versioned+: Use versioning on the bucket.
      class Settings
        def initialize(settings = {})
          clear
          update(settings)
        end

        def update(settings)
          case settings
          # Old style setup allowed settings to be an array of allowed access
          # keys.
          when Set, Array
            allow_access_keys(settings)
          else
            settings.each do |name, value|
              public_send("#{name}", value)
            end
          end
        end

        def access_keys
          @access_keys
        end

        def replicates_to
          @replicates_to
        end

        def versioned?
          !!@versioned
        end

        def allow_access_keys(access_key_ids)
          @access_keys.merge(access_key_ids)
        end

        def replicated?
          @replicates_to && !@replicates_to.empty?
        end

        def replicate_to(replicates_to)
          @replicates_to = replicates_to
        end

        def replicate_to_buckets
          return [] unless replicated?
          @replicates_to.map do |location|
            if bucket = Canistor.store.dig(*location.split(':'))
              bucket
            else
              raise(
                Canistor::ConfigurationError,
                "Can't locate bucket `#{location}' when trying to replicate " \
                "object. Please make sure the replication location is in the " \
                "form of region:bucket and configured in Canistor."
              )
            end
          end
        end

        def versioned(versioned)
          @versioned = versioned
        end

        def clear
          @access_keys = Set.new
          @replicates_to = []
          @versioned = false
        end
      end
    end
  end
end
