# frozen_string_literal: true

require "digest/md5"

module Canistor
  module Storage
    class Object
      # Renders the result of a copy request.
      class CopyObjectResult
        def initialize(source_object, object)
          @source_object, @object = source_object, object
        end

        def headers
          @object.identity_headers.merge(
            'x-amz-copy-source-version-id' => @source_object.version_id
          )
        end

        def to_xml
          Nokogiri::XML::Builder.new do |xml|
            xml.CopyObjectResult do
              xml.LastModified @object.last_modified
              xml.ETag @object.etag
            end
          end.to_xml
        end

        def self.serve(context, subject, source_object, object)
          result = new(source_object, object)
          context.http_response.signal_headers(200, result.headers)
          context.http_response.signal_data(result.to_xml)
        end
      end

      attr_accessor :region
      attr_accessor :bucket
      attr_accessor :key

      attr_reader :data
      attr_reader :last_modified

      def initialize(**attributes)
        @versioned = false
        @headers = {}
        @data = ''
        attributes.each do |name, value|
          send("#{name}=", value)
        end
        @digest = nil
      end

      def versioned=(versioned)
        @versioned = versioned
      end

      def versioned?
        !!@versioned
      end

      def version_id
        if versioned?
          @version_id ||= generate_version_id
        end
      end

      def size
        data&.size
      end

      def label
        [region, bucket, key].map(&:to_s).join(':') + ' ' + headers.inspect
      end

      def write(headers, data)
        self.headers = headers
        self.data = data
      end

      def digest
        @digest ||= Digest::MD5.hexdigest(data)
      end

      def etag
        '"' + digest + '"'
      end

      def endpoint
        Aws::Partitions::EndpointProvider.resolve(region, 's3')
      end

      def location
        [
          endpoint,
          bucket,
          key,
        ].join('/')
      end

      def headers
        @headers.merge(identity_headers).merge(
          'date' => Time.now.httpdate,
          'content-length' => size.to_s,
          'last-modified' => last_modified.httpdate,
          'server' => 'Canistor'
        )
      end

      def identity_headers
        {
          'etag' => etag,
          'x-amz-id' => digest[0, 16],
          'x-amz-id-2' => Base64.strict_encode64(digest),
          'x-amz-request-id' => Base64.strict_encode64(SecureRandom.hex(16)),
          'x-amz-version-id' => version_id
        }.compact
      end

      def head(context, subject)
        context.http_response.signal_headers(200, headers)
      end

      def get(context, subject)
        context.http_response.signal_headers(200, headers)
        context.http_response.signal_data(data)
      end

      def put(context, subject)
        if source_object = find_source(context, subject)
          apply_source_object(context, subject, source_object)
        else
          apply_request(context, subject)
        end
      end

      def copy
        object = self.class.new(region: region, bucket: bucket, key: key)
        object.write(headers, data)
        object
      end

      def delete(context, subject)
        context.http_response.signal_headers(200, {})
      end

      def copy_headers(context)
        directive = context.http_request.headers['x-amz-metadata-directive']
        case directive
        when 'COPY'
          identity_headers
        when 'REPLACE', nil
          context.http_request.headers
        else
          raise ArgumentError, "Unsupported metadata directive: `#{directive}'"
        end
      end

      private

      def find_source(context, subject)
        if source = context.http_request.headers['x-amz-copy-source']
          bucket_name, key = source.split('/', 2)
          if bucket = Canistor.find_bucket(region, bucket_name)
            if object = bucket.dig(key)
              object
            else
              Canistor::ErrorHandler.serve_no_such_key(context, subject)
              throw :rendered_error
            end
          else
            Canistor::ErrorHandler.serve_no_such_bucket(context, subject)
            throw :rendered_error
          end
        end
      end

      def apply_request(context, subject)
        write(context.http_request.headers, context.http_request.body)
        context.http_response.signal_headers(200, identity_headers)
      end

      def apply_source_object(context, subject, source_object)
        write(source_object.copy_headers(context), source_object.data)
        CopyObjectResult.serve(context, subject, source_object, self)
      end

      def generate_version_id
        Base64.strict_encode64(SecureRandom.hex(16)).gsub('=', '')
      end

      META_HEADERS = %w(
        content-disposition
        content-type
      )

      def headers=(headers)
        return if headers.nil?
        headers.each do |name, value|
          if META_HEADERS.include?(name) || name.start_with?('x-amz-meta')
            @headers[name] = value
          end
        end
      end

      def data=(data)
        @digest = nil
        @last_modified = Time.now
        if data.respond_to?(:read)
          @data = data.read
        else
          @data = data
        end
      end
    end
  end
end
